# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#

import logging
from .helpers import json_formatter
from .helpers import loop_objects_names
from .helpers import find_csv_objects
from .connect import Connection

class ObjectIterator:
    
    def __init__(self, connection, cos_bucket):
        self.connection = connection
        self.cos_bucket = cos_bucket

    def _paginator_iterator(self):
        """ use the paginate function to iterate through all the objects in a bucket """

        cos = self.connection
        paginator = cos.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(Bucket=self.cos_bucket)
                
        return page_iterator
    
    def _get_objects(self):
        """ get all the objects from the paginator and return as a list of objects """
        
        page_iterator = self._paginator_iterator()
        bucket_list = []
        for page in page_iterator:
            if "Contents" in page:
                if any(bucket_list) == False:
                    bucket_list = page["Contents"]
                else:
                    for key in page['Contents']:
                        bucket_list.append(key)
            else:
                logging.error("No objects are stored in the bucket!")

        return bucket_list
    
    def _get_csv_objects(self):
        """ get only the objects that are CSV object """
        objs = self._get_objects()

        return find_csv_objects(objs)

    def list_all_objects(self):
        """ pretty print all of the objects - even non-CSV types """
        try:
            objs = self._get_objects()

            return json_formatter(objs)

        except Exception as e:
            logging.error(e)

    def list_csv_objects(self):
        """ pretty print all of the CSV objects """
        try:
            csv_files = self._get_csv_objects()

            return json_formatter(csv_files)

        except Exception as e:
            logging.error(e)
    
    def list_csv_object_names(self):
        """ list only the names of the objects, not the other metadata """
        try:
            csv_files = self._get_csv_objects()

            return loop_objects_names(csv_files)
                
        except Exception as e:
            logging.error(e)

def object_iterator(connection, cos_bucket):
        """A helper to create a Bucket object."""

        return ObjectIterator(connection, cos_bucket)



# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#
from .convert_module import convert_objects

def flags(args):
    """
    Starts the program to convert CSV objects to Parquet.
    
    Required flags:
    -a or --apikey      -   IBM Cloud API Key
    -e or --endpoint    -   URI endpoint of the COS bucket being used
    -b or --bucket      -   COS bucket that will be used to check for objects, convert then, and place them back into the bucket

    Future TODO: Users can select a different output bucket as long as they are have the same COS endpoint
    """

    converter = convert_objects(args.apikey, args.cos_endpoint, args.cos_bucket)

    if args.list_all:
        """ 
        -l - Lists all the files in a COS bucket
        """
        print(converter.list_bucket_objects())
        return

    if args.csv:
        """ 
        -c - Lists only CSV files in a COS bucket
        """
        print(converter.list_csv_objects())
        return
    
    if args.csv_names:
        """
        -cn - Lists only the names of CSV objects in a COS bucket
        """
        for i, name in enumerate(converter.list_csv_objects_names()):
            print("{}) {}".format((i + 1), name))

        return

    if args.files:
        """ 
        Flags -f and -n must be used together and the number of CSV objects names must match the new Parquet object names.

        -f - Names CSV objects to convert. If you attempt to name the same object twice,
            it will only accept the name once.
        -n - New names of converted objects to Parquet. Names can include prefixes like `my/new/object.parquet`. 
            Do not put the `.parquet` file name after the new name. It's already added for you. The number of new names must
            equal the number of object names you want converted or it will not convert the objects.
        """
        
        csvs_to_convert = args.files
        csvs_in_bucket = converter.list_csv_objects_names()
        objs_new_names = args.names

        if len(csvs_to_convert) != len(objs_new_names):
            print("The number of objects to convert doesn't equal the new names you provided.")
            return

        if not objs_new_names:
            print("You need to provide new name(s) for the objects you want to convert.")
            return

        objs_to_convert = [obj for obj in csvs_to_convert if obj in csvs_in_bucket]
        
        for converted_obj, name in zip(objs_to_convert, objs_new_names):
            print("Now Converting: {} --> {}.parquet".format(converted_obj, name))
        
        converter.convert_selected_objects(objs_to_convert, objs_new_names)
    
    elif not args.files:
        print("You need to provide the names of objects you want to convert.")



# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#
#!/usr/bin/env python3
import argparse
from .cli_commands import flags

def main():
    parser = argparse.ArgumentParser(description='Convert CSV Objects to Parquet')

    parser.add_argument('-a', '--apikey', help='IBM Cloud API Key', required=True)
    parser.add_argument('-e', '--cos-endpoint', help='IBM COS Endpoint URL', required=True)
    parser.add_argument('-b', '--cos-bucket', help='IBM COS Bucket Name', required=True)
    parser.add_argument('-l', '--list-all', action='store_true', help='List All Objects in IBM COS Bucket')
    parser.add_argument('-c', '--csv', action='store_true', help='List All CSV Objects in IBM COS Bucket')
    parser.add_argument('-f', '--files', nargs="+", help='Convert Selected Objects in IBM COS Bucket - must use with -n')
    parser.add_argument('-n', '--names', nargs="+", help='Name(s) of the New Parquet Objects - must use with -f')
    parser.add_argument('-cn', '--csv-names', action="store_true", help='Only the names of all CSV objects')

    args = parser.parse_args()

    flags(args)

if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#

import logging
import pandas as pd
from io import BytesIO
from io import StringIO
from .connect import Connection
from .object_module import object_iterator

class Converter:
    """Class to Convert CSV to Parquet"""

    def __init__(self, connection, cos_bucket):
        """ initialize the class with the connection to COS and COS bucket """

        self.connection = connection
        self.cos_bucket = cos_bucket
        self.object_iterator = object_iterator(connection, cos_bucket)

    def _get_object(self, obj):
        """ get the body of an object """

        cos = self.connection
        cos_object = cos.get_object(Bucket=self.cos_bucket, Key=obj)
        body = cos_object['Body'].read().decode('utf-8')

        return body

    def _read_csv_to_df(self, obj):
        """ read the object contents and put them into a panda's dataframe """

        get_object = self._get_object(obj)
        buff = StringIO(get_object)

        try:
            df = pd.read_csv(buff)

            # replaces spaces in column names like `sample column` with `sample_column`
            cols = df.columns.str.strip().str.replace(' ', '_')
            df.columns = cols   

            return df

        except Exception:
            logging.error("Can't create panda's dataframe from object")

    def _convert_to_parquet(self, obj, new_obj_name):
        """ convert the pandas dataframe to Parquet """

        df = self._read_csv_to_df(obj)
        parquet_obj = BytesIO()
        df.to_parquet(parquet_obj, compression="gzip", engine="pyarrow")
        parquet_obj.seek(0)
        
        return self._save_object(new_obj_name, parquet_obj)
    
    def _save_object(self, new_obj_name, parquet_data):
        """ save the body of the object and rename with the .parquet file extension """

        cos = self.connection
        new_obj_name_parquet = "{}.parquet".format(new_obj_name)
        return cos.put_object(Body=parquet_data.getvalue(), Bucket=self.cos_bucket, Key=new_obj_name_parquet)

    def list_bucket_objects(self):
        """ list all the objects in a bucket """
        
        return self.object_iterator.list_all_objects()

    def list_csv_objects(self):
        """ list all the objects that are CSV objects in a bucket """

        return self.object_iterator.list_csv_objects()

    def list_csv_objects_names(self):
        """ list only the names of CSV objects """
        
        return self.object_iterator.list_csv_object_names()

    def convert_selected_objects(self, objs, names):
        """ convert selected objects to Parquet """

        for csv_name, new_parquet_name in zip(objs, names):
            self._convert_to_parquet(csv_name, new_parquet_name)
            
def convert_objects(api_key, service_endpoint, cos_bucket):
        """ a helper to create a convert object """

        connection = Connection(api_key, service_endpoint)
        return Converter(connection, cos_bucket)



# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='csvtoparquet',
    version='0.1.4',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    install_requires=['pandas','pyarrow','ibm-cos-sdk','simplejson'],
    author='IBM',
    author_email='abdullah.alger@ibm.com',
    license='Apache 2.0',
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
        'Natural Language :: English'
    ],
    keywords='cloud csv parquet object_storage IBM',
    packages=['csvtoparquet'],
    entry_points={
        'console_scripts': ['csvtoparquet = csvtoparquet.converter:main']
    }
)


# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#

import logging
import datetime
import simplejson as json
import csv
import os
    
def _datetime_helper(objs):
    """ helper function for datetime in JSON reponse """

    if isinstance(objs, datetime.datetime):
        return objs.isoformat()
    raise TypeError("Unknown Type")

def json_formatter(json_file):
    """ helper function to format JSON results """

    return json.dumps(json_file, default=_datetime_helper, indent=4 * ' ')

def find_csv_objects(objs):
    """ helper function to find only CSV files """

    return [obj for obj in objs if '.csv' in obj.get('Key')]
    
def loop_objects_names(objs):
    """ helper function to loop through list of objects """

    return [obj['Key'] for obj in objs]

def object_convert_check():
    """ TODO """
    pass

def display_conversion():
    """ TODO """
    pass

# def _split_object_prefix_name(obj):
#     """ helper function that returns a tuple with the bucket prefix and object name """
    
#     return os.path.split(obj)

# def remove_objects_extensions(obj):
#     """ helper function to remove .csv from object name """

#     obj_name =  _split_object_prefix_name(obj)[1]
#     return os.path.splitext(obj_name)[0]

# def get_bucket_prefix(obj):
#     """ helper function to separate bucket prefix from object """

#     obj_prefix = _split_object_prefix_name(obj)[0]
#     return '/' if '' == obj_prefix else "{}/".format(obj_prefix)



# ------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
#
import logging
import ibm_boto3
from ibm_botocore.client import Config

def Connection(api_key, service_endpoint):
    """ create a connection to COS """

    try:
        cos = ibm_boto3.client('s3',
                    ibm_api_key_id=api_key,
                    ibm_auth_endpoint='https://iam.bluemix.net/oidc/token',
                    config=Config(signature_version='oauth'),
                    endpoint_url="https://{}".format(service_endpoint))

        return cos

    except Exception as e:
        logging.error(e)